/**
 * @license
 * Copyright 2023 Google LLC
 * SPDX-License-Identifier: Apache-2.0
 */
/**
 * A logger for watch mode that avoids useless output.
 */
export class WatchLogger {
    constructor(actualLogger) {
        this._iterationBuffer = [];
        this._iterationIsInteresting = 
        /* The first iteration is always interesting. */ true;
        this._actualLogger = actualLogger;
    }
    log(event) {
        if (this._iterationIsInteresting) {
            // This iteration previously had an interesting event (or it's the very
            // first one, which we always show).
            this._actualLogger.log(event);
            if (this._isWatchRunEnd(event)) {
                this._iterationIsInteresting = false;
            }
        }
        else if (this._isWatchRunEnd(event)) {
            // We finished a watch iteration and nothing interesting ever happened.
            // Discard the buffer.
            this._iterationBuffer.length = 0;
        }
        else if (this._isInteresting(event)) {
            // The first interesting event of the iteration. Flush the buffer and log
            // everything from now until the next iteration.
            while (this._iterationBuffer.length > 0) {
                this._actualLogger.log(this._iterationBuffer.shift());
            }
            this._actualLogger.log(event);
            this._iterationIsInteresting = true;
        }
        else {
            // An uninteresting event in a thus far uninteresting iteration.
            this._iterationBuffer.push(event);
        }
    }
    _isInteresting(event) {
        const code = event.type === 'output'
            ? event.stream
            : event.type === 'info'
                ? event.detail
                : event.reason;
        switch (code) {
            case 'fresh':
            case 'no-command':
            case 'failed-previous-watch-iteration':
            case 'watch-run-start':
            case 'start-cancelled':
            case 'locked': {
                return false;
            }
        }
        return true;
    }
    _isWatchRunEnd(event) {
        return event.type === 'info' && event.detail === 'watch-run-end';
    }
}
//# sourceMappingURL=watch-logger.js.map